# Contributing to Drupal commercetools module

We recommend using [DDEV](https://ddev.com/) for local development and testing,
but any local development environment that meets Drupal 10 system requirements
and has Composer installed will do.

The module is covered by unit and functional tests to simplify regression
testing, using PHPUnit and Nightwatch.

So please run tests locally to check that your changes don't break anything,
and adapt tests if you change anything, that already covered by the current
tests.

## Quick setup of the local environment using DDEV

```
# Create a new DDEV project
ddev config --docroot=web --project-type=drupal10
# Add support for running Nightwatch tests using Selenium
ddev get ddev/ddev-selenium-standalone-chrome
ddev composer create drupal/recommended-project --no-install
ddev composer install
ddev composer require drush/drush
ddev composer require drupal/core-dev --dev --update-with-all-dependencies
ddev drush site-install

# Install the module
ddev composer require drupal/commercetools

# Log in into the container
ddev ssh
```

## commercetools Content submodule

The module renders the commercetools contend on the backend using Drupal API
and deliver it to the frontend as Drupal html pages.


## commercetools Decoupled submodule

The module provides decoupled blocks for the commercetools, which load
and render the content on the browser side.



## commercetools API documentation

The module uses the commercetools SDK API, mostly the GraphQL API.

Documentation references:

- [PHP SDK](https://docs.commercetools.com/sdk/php-sdk)

- [JavaScript SDK](https://docs.commercetools.com/sdk/typescript-sdk)

- [REST API documentation](https://docs.commercetools.com/learning-composable-commerce-developer-essentials/api-queries-query-predicates/overview)

- [GraphQL documentation](https://docs.commercetools.com/learning-composable-commerce-developer-essentials/graphql/overview)

- [GraphQL playground](https://impex.us-central1.gcp.commercetools.com/graphiql)

- [Constructing the "where" query parameter](https://docs.commercetools.com/api/predicates/query)


## Running tests

The module uses commercetools API, so for almost each test we need correct
responses from the commercetools API. And for that you need the working
commercetools credentials!

To workaround this, the tests store the real responses from the API
in files and mock them when running tests. The responses content is stored in
the directory `tests/modules/commercetools_test/assets`.

If you write new tests or modify existing ones that modifies the commercetools
API requests (for example, add new fields to the GraphQL query), you should use
the real credentials to update the asset files. To make this work, use the
environment values to control the tests running mode:

- `CT_USE_REAL_API=1` - this variable enables making real requests to
  commercetools and storing responses to files.
- `CT_CLIENT_ID=xxx` - commercetools Client ID.
- `CT_CLIENT_SECRET=yyy` - commercetools Client Secret.

With these variables, the tests will do real requests to commercetools and store
them into assets, to make the tests run well next time without any credentials.

To run Nightwatch tests locally you need to install additional dependencies from
the `require-dev` section of the module's `composer.json` file and apply patches.

### Run PHPUnit tests

```
cd web/core && phpunit ../web/modules/contrib/commercetools
```

### Run Nightwatch tests (End-to-end browser tests)

#### Install Nightwatch dependencies (only once)

```
cd web/core && yarn
cd web/modules/contrib/commercetools && yarn
```

#### Run Nightwatch tests

To run all Nightwatch tests - use this command:
```
cd web/core && yarn test:nightwatch --tag commercetools
```
or to run a single test - put the full path to the test like this
```
cd web/core && yarn test:nightwatch ../modules/contrib/commercetools/modules/commercetools_demo/tests/src/Nightwatch/Tests/demoContentSmokeTest.js
```

### Upgrade to Nightwatch 3.x for Drupal below 11.0.

We use Nightwatch 3.x version because 2.x has issues with finding nested
elements (`element('selector1').findElement('selector2')`).

To install Nightwatch 3 on your local machine, just open the `web/core`
directory and run:
```
yarn add --dev nightwatch@^3
```

Also, Nightwatch from version 2.5+ requires the more recent version of Selenium
container, so instead of the `seleniarm/standalone-chromium:4.1.4-20220429` -
use the image `selenium/standalone-chrome:latest` and enable W3C compatibility
by the env: `DRUPAL_TEST_WEBDRIVER_W3C=1`.

##### Run a single test on the pipeline.

To run a single specific test on the pipeline, without waiting for other tests
and steps, you can use this approach:

1. Comment the general tag for all tests in the file
`tests/src/Nightwatch/Lib/getBeforeAfterFunctions.js`:
```diff
-   const testTags = ['commercetools', moduleName];
+   // const testTags = ['commercetools', moduleName];
```

2. Override this value in the single test file that you want to run:
```diff
  ...getBeforeAfterFunctions(__filename),
+    '@tags': ['commercetools'],
```

3. Disable unnecessary steps in the pipeline by temporary extending the
`.gitlab-ci.yml` file:
```yaml
variables:
  # ...
  SKIP_COMPOSER_LINT: "1"
  SKIP_PAGES: "1"
  SKIP_PHPCS: "1"
  SKIP_PHPSTAN: "1"
  SKIP_PHPUNIT: "1"
  SKIP_STYLELINT: "1"
  SKIP_ESLINT: "1"
  SKIP_CSPELL: "1"
  # SKIP_NIGHTWATCH: "1"
  OPT_IN_TEST_MAX_PHP: 0"
  OPT_IN_TEST_PREVIOUS_MAJOR: "0"
```

## Code Style

Follow the general code style rules from Drupal.org, and in addition to this:

- For any string variable try to use constants as much as possible, to simplify
finding the dependencies of the string and simplify changing the values.

### Hooks

Do not put any business logic into hook_* functions. Write all actual code as
static functions in classes like CommercetoolsUpdates::installCustomerIdField(),
and call them from the hook. This is required to simplify covering the code
by tests and the code re-usage.

So, keep in the hook only the data definitions, calls of functions and messages
output only. See the previous hook updates as examples.

Also, do not use constants in the `hook_update_*` functions, because the value of
the constants can be changed it time, that will break the old hook updates.

## Tips and Tricks

### Missing stored responses

If you face an exception like this when running a test:
```
Exception: Missing the stored response file for the request with hash [md5 hash]
```
this means that you modified in the code a GraphQL request, executed during this 
test run, and you need to update the mocked response for this GraphQL request.

This can be done by re-running the test with the `CT_USE_REAL_API=1` environment
variable, and adding to git the files with the body of the responses from API.
