<?php

namespace Drupal\commercetools_content\Form;

use Drupal\commercetools\CommercetoolsCarts;
use Drupal\commercetools\Form\UiModuleSettingsFormBase;
use Drupal\commercetools\Routing\UiModulesRouteProviderBase;
use Drupal\commercetools_content\Routing\RouteProvider;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides cart form.
 */
class CartForm extends FormBase {


  /**
   * The commercetools cart service.
   *
   * @var \Drupal\commercetools\CommercetoolsCarts
   */
  protected CommercetoolsCarts $ctCart;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);
    $instance->ctCart = $container->get('commercetools.carts');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'commercetools_content_cart_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $install_state = NULL) {
    $cartResponse = $this->ctCart->getCurrentCart();
    $cart = !empty($cartResponse)
      ? $cartResponse->getData()
      : [
        'totalLineItemQuantity' => 0,
        'lineItems' => [],
      ];

    $form['#theme'] = 'commercetools_cart_form';
    $form['#attributes']['id'][] = 'commercetools-cart-form';
    $form['#attached']['library'][] = 'commercetools_content/commercetools_content.carts';
    $form['#cart'] = $cart;

    $form['lineItems'] = [
      '#type' => 'container',
      '#tree' => TRUE,
    ];
    foreach ($cart['lineItems'] as $lineItem) {
      $form['lineItems'][$lineItem['id']]['quantity'] = [
        '#type' => 'hidden',
        '#default_value' => $lineItem['quantity'],
        '#attributes' => ['class' => ['line-item-quantity']],
      ];
      $form['lineItems'][$lineItem['id']]['remove'] = [
        '#type' => 'hidden',
        '#default_value' => FALSE,
        '#attributes' => ['class' => ['line-item-quantity']],
      ];
    }

    $form['cartUpdate'] = [
      '#type' => 'submit',
      '#value' => $this->t('Update'),
      '#submit' => ['::updateCart'],
      '#ajax' => [
        'callback' => '::ajaxCallback',
        'disable-refocus' => FALSE,
        'event' => 'click',
        'wrapper' => 'commercetools-cart-form',
        'progress' => [
          'type' => 'fullscreen',
          'message' => $this->t('Updating...'),
        ],
      ],
      '#attributes' => ['class' => ['cart-update-button']],
    ];

    $checkoutPath = $this->config(ContentSettingsForm::CONFIGURATION_NAME)
      ->get(UiModuleSettingsFormBase::CONFIG_CHECKOUT_PATH);

    $form['actions'] = [
      '#type' => 'actions',
      '#access' => !empty($cart['lineItems']),
      'submit' => [
        '#type' => 'link',
        '#title' => $this->t('Order'),
        '#url' => isset($checkoutPath)
          ? Url::fromRoute(RouteProvider::ROUTE_PREFIX . UiModulesRouteProviderBase::PAGE_CHECKOUT_ROUTE)
          : Url::fromUserInput('#'),
        '#attributes' => [
          'class' => [
            'btn',
            'btn-primary',
            isset($checkoutPath) ? '' : 'disabled',
          ],
        ],
      ],
    ];

    $form['#cache']['contexts'] = [
      'commercetools_cart',
    ];
    return $form;
  }

  /**
   * Ajax callback to update product variant.
   */
  public function ajaxCallback(array &$form, FormStateInterface $form_state) {
    return $form;
  }

  /**
   * Submit function to update a user cart.
   */
  public function updateCart(array $form, FormStateInterface $form_state) {
    $cart = $this->ctCart->getCurrentCart()->getData();
    $actions = [];
    foreach ($form_state->getValue('lineItems') as $lineItemId => $lineItem) {
      $actions[] = $lineItem['remove']
        ? ['removeLineItem' => ['lineItemId' => $lineItemId]]
        : ['changeLineItemQuantity' => ['lineItemId' => $lineItemId, 'quantity' => (int) $lineItem['quantity']]];
    }
    $this->ctCart->updateCart($cart['id'], $cart['version'], $actions);
    $form_state->setRebuild(TRUE);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
  }

}
