<?php

namespace Drupal\commercetools_content\Form;

use Drupal\commercetools\CommercetoolsService;
use Drupal\commercetools_content\Service\CommercetoolsContentComponents;
use Drupal\Core\Form\BaseFormIdInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides catalog facets.
 *
 * @internal
 */
class CatalogFacetsForm extends FormBase implements BaseFormIdInterface {

  /**
   * {@inheritdoc}
   */
  public function getBaseFormId() {
    return 'commercetools_content_catalog_facets_form';
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    // Try to fix an error with several identical forms on the same page.
    // @todo Remove after https://www.drupal.org/project/drupal/issues/2821852
    static $index = 0;
    return $this->getBaseFormId() . $index++;
  }

  /**
   * The Commercetools service.
   *
   * @var \Drupal\commercetools\CommercetoolsService
   */
  protected CommercetoolsService $ct;

  /**
   * The renderer service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected RendererInterface $renderer;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);
    $instance->ct = $container->get('commercetools');
    $instance->renderer = $container->get('renderer');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $install_state = NULL) {
    [$facets, $activeFilters, $productListIndex] = $form_state->getBuildInfo()['args'];
    $paramName = CommercetoolsContentComponents::addParameterSuffix('facets', $productListIndex);

    $form['#attributes']['data-product-list-index'] = $productListIndex;
    $form['#attributes']['class'][] = 'commercetools-content-auto-submit';
    $form['#attached'] = ['library' => ['commercetools_content/commercetools_content.auto_submit']];
    $form[$paramName] = [
      '#type' => 'container',
      '#tree' => TRUE,
    ];
    foreach ($facets as $facet) {
      $options = [];
      foreach ($facet['value']['terms'] as $term) {
        $labelRender = [
          '#theme' => 'commercetools_filter_option',
          '#label' => $term['label'] ?? $term['term'],
          '#count' => $term['productCount'],
        ];
        $options[$term['term']] = $this->renderer->renderRoot($labelRender);
      }
      if (!empty($options)) {
        $label = [
          '#theme' => 'commercetools_filter_label',
          '#label' => $facet['label'] ?? $facet['facet'],
        ];
        $form[$paramName][$facet['facet']] = [
          '#title' => $this->renderer->renderRoot($label),
          '#type' => 'checkboxes',
          '#options' => $options,
          '#default_value' => $activeFilters[$facet['facet']] ?? [],
        ];
      }
    }

    $form['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Update'),
      '#attributes' => ['class' => ['js-hidden']],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $query = $this->getRequest()->query->all();
    // Reset page after applying new facets.
    unset($query['page']);

    [, , $productListIndex, $targetPage] = $form_state->getBuildInfo()['args'];
    $paramName = CommercetoolsContentComponents::addParameterSuffix('facets', $productListIndex);
    $query[$paramName] = [];
    foreach ($form_state->getValue($paramName) as $facetName => $facetValues) {
      $facetValues = array_filter($facetValues);
      if (!empty($facetValues)) {
        $query[$paramName][$facetName] = $facetValues;
      }
    }

    $targetPage ??= Url::fromRoute('<current>');
    $targetPage->setOption('query', array_filter($query));
    $form_state->setRedirectUrl($targetPage);
  }

  /**
   * Initialize the form state and the entity before the first form build.
   */
  protected function init(FormStateInterface $form_state) {

  }

}
