<?php

namespace Drupal\commercetools_content\Form;

use Drupal\commercetools\CommercetoolsApiServiceInterface;
use Drupal\commercetools\CommercetoolsCarts;
use Drupal\commercetools\CommercetoolsLocalization;
use Drupal\commercetools\CommercetoolsService;
use Drupal\commercetools\Routing\UiModulesRouteProviderBase;
use Drupal\commercetools_content\Routing\RouteProvider;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides order submission form.
 */
class OrderSubmissionForm extends FormBase {

  /**
   * The Commercetools service.
   *
   * @var \Drupal\commercetools\CommercetoolsService
   */
  protected CommercetoolsService $ct;
  /**
   * The Commercetools API service.
   *
   * @var \Drupal\commercetools\CommercetoolsApiServiceInterface
   */
  protected CommercetoolsApiServiceInterface $ctApi;

  /**
   * The commercetools cart service.
   *
   * @var \Drupal\commercetools\CommercetoolsCarts
   */
  protected CommercetoolsCarts $ctCart;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);
    $instance->ctApi = $container->get('commercetools.api');
    $instance->ct = $container->get('commercetools');
    $instance->ctCart = $container->get('commercetools.carts');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'commercetools_content_order_submission';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $install_state = NULL) {
    $settings = $this->config(CommercetoolsService::CONFIGURATION_NAME);
    $cart = $this->ctCart->getCurrentCart()->getData();

    $addressFields = [
      'shippingAddress' => $this->t('Shipping address'),
      'billingAddress' => $this->t('Billing address'),
    ];

    foreach ($addressFields as $key => $label) {
      $form[$key] = [
        '#type' => 'fieldset',
        '#tree' => TRUE,
        '#weight' => 1,
        '#attributes' => [
          'class' => ['shadow', 'p-3', 'bg-primary', 'bg-opacity-10'],
        ],
        'label' => [
          '#type' => 'html_tag',
          '#tag' => 'h2',
          '#value' => $label,
          '#attributes' => [
            'class' => ['pb-3', 'mb-4', 'border-bottom'],
          ],
        ],
        'firstName' => [
          '#type' => 'textfield',
          '#title' => $this->t('First name'),
          '#default_value' => $cart[$key]['firstName'] ?? NULL,
          '#required' => TRUE,
        ],
        'lastName' => [
          '#type' => 'textfield',
          '#title' => $this->t('Last name'),
          '#default_value' => $cart[$key]['lastName'] ?? NULL,
          '#required' => TRUE,
        ],
        'email' => [
          '#type' => 'email',
          '#title' => $this->t('Email'),
          '#default_value' => $cart[$key]['email'] ?? $this->currentUser()->getEmail() ?? NULL,
          '#disabled' => $this->currentUser()->isAuthenticated(),
          '#required' => TRUE,
        ],
        'country' => [
          '#type' => 'textfield',
          '#title' => $this->t('Country'),
          '#default_value' => $settings->get(CommercetoolsLocalization::CONFIG_COUNTRY),
          '#disabled' => TRUE,
        ],
        'streetName' => [
          '#type' => 'textfield',
          '#title' => $this->t('Street Address'),
          '#default_value' => $cart[$key]['streetName'] ?? NULL,
          '#required' => TRUE,
        ],
        'streetNumber' => [
          '#type' => 'textfield',
          '#title' => $this->t('Street number'),
          '#default_value' => $cart[$key]['streetNumber'] ?? NULL,
        ],
        'state' => [
          '#type' => 'textfield',
          '#title' => $this->t('State'),
          '#default_value' => $cart[$key]['state'] ?? NULL,
          '#required' => TRUE,
        ],
        'city' => [
          '#type' => 'textfield',
          '#title' => $this->t('City'),
          '#default_value' => $cart[$key]['city'] ?? NULL,
          '#required' => TRUE,
        ],
        'postalCode' => [
          '#type' => 'textfield',
          '#title' => $this->t('ZIP code'),
          '#default_value' => $cart[$key]['postalCode'] ?? NULL,
          '#required' => TRUE,
        ],
        'phone' => [
          '#type' => 'textfield',
          '#title' => $this->t('Phone number'),
          '#default_value' => $cart[$key]['phone'] ?? NULL,
        ],
      ];
    }

    $form['same_address'] = [
      '#prefix' => '<div class="mt-4 mb-4 shadow p-3 bg-primary bg-opacity-10">',
      '#suffix' => '</div>',
      '#type' => 'checkbox',
      '#title' => $this->t('Billing address same as shipping address'),
      '#default_value' => empty($cart['billingAddress']),
      '#weight' => 2,
    ];

    $form['billingAddress']['#weight'] = 3;
    foreach ($form['billingAddress'] as &$element) {
      if (isset($element['#required'])) {
        $element['#required'] = FALSE;
        $element['#states'] = [
          'required' => [
            ':input[name="same_address"]' => ['checked' => FALSE],
          ],
        ];
      }
    }
    $form['billingAddress']['#states'] = [
      'invisible' => [
        ':input[name="same_address"]' => ['checked' => TRUE],
      ],
    ];

    $form['actions'] = [
      '#type' => 'actions',
      'submit' => [
        '#type' => 'submit',
        '#value' => $this->t('Save & Continue'),
      ],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $cart = $this->ctCart->getCurrentCart()->getData();
    $shippingAddress = array_filter($form_state->getValue('shippingAddress'));
    $billingAddress = $form_state->getValue('same_address') ? $shippingAddress : array_filter($form_state->getValue('billingAddress'));

    $response = $this->ctCart->updateCart($cart['id'], $cart['version'], [
      [
        'setShippingAddress' => [
          'address' => $shippingAddress,
        ],
      ],
      [
        'setBillingAddress' => [
          'address' => $billingAddress,
        ],
      ],
    ]);

    $order = $this->ctCart->createOrderFromCart($response->getData())->getData();
    $this->ctCart->deleteCurrentCart();
    $form_state->setRedirect(RouteProvider::ROUTE_PREFIX . UiModulesRouteProviderBase::PAGE_USER_ORDER_ROUTE, [
      'user' => $this->currentUser()->id(),
      'orderId' => $order['id'],
    ]);
  }

}
