<?php

namespace Drupal\commercetools_content\Plugin\Block;

use Drupal\commercetools\Plugin\Block\CommercetoolsCategoriesListBlockBase;
use Drupal\commercetools_content\Form\ContentSettingsForm;
use Drupal\commercetools_content\Service\CommercetoolsContentComponents;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Template\Attribute;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a Categories List block.
 *
 * @Block(
 *   id = "commercetools_content_categories_list",
 *   admin_label = @Translation("Categories List"),
 * )
 */
class CommercetoolsContentCategoriesListBlock extends CommercetoolsCategoriesListBlockBase {

  /**
   * The commercetools content component service.
   *
   * @var \Drupal\commercetools_content\Service\CommercetoolsContentComponents
   */
  protected $contentComponents;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): self {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->contentComponents = $container->get('commercetools_content.content_components');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function build(): array {
    $build = [];

    $treeDepth = $this->configuration['max_level'];
    $index = $this->configuration['product_list_index'];
    $queryParams = $this->contentComponents->getRequestQueryParams();
    $paramName = CommercetoolsContentComponents::addParameterSuffix('category', $index);
    $categoriesTree = $this->getCategoriesTree($queryParams[$paramName] ?? NULL);
    if ($categoriesTree) {
      // Prevent re-ordering by Drupal\Core\Render\Element::children().
      $build['#sorted'] = TRUE;
      // Add the theme wrapper for outer markup and allow theme overrides.
      $build['#theme'] = 'menu__' . $this->getPluginId();
      $build['#items'] = $this->buildItems(
        $categoriesTree,
        $index,
        $queryParams,
        $treeDepth,
      );
    }

    return $build;
  }

  /**
   * Builds the #items property for a categories tree.
   *
   * @param array $list
   *   The data structure representing the categories tree.
   * @param int $index
   *   The component index.
   * @param array $queryParams
   *   The current URL query parameters.
   * @param int $treeDepth
   *   The maximum allowed categories tree depth.
   * @param int $currentDepth
   *   The current level. Optional.
   *
   * @return array
   *   A value to use for the #items property to render a system menu.
   */
  protected function buildItems(array $list, int $index, array $queryParams, int $treeDepth, int $currentDepth = 1): array {
    $items = [];

    $categoryParameterName = CommercetoolsContentComponents::addParameterSuffix('category', $index);
    $baseLink = $this->getLink();
    foreach ($list as $data) {
      $link = clone $baseLink;
      $queryParams[$categoryParameterName] = $data['id'];
      $link->setOption('query', $queryParams);
      $element = [
        'in_active_trail' => !empty($data['in_active_trail']),
        'attributes' => new Attribute(),
        'title' => $data['name'],
        'url' => $link,
        'below' => [],
      ];

      if ($data['children'] && (!empty($data['in_active_trail']) || !$treeDepth || $currentDepth < $treeDepth)) {
        $element['below'] = $this->buildItems(
          $data['children'],
          $index,
          $queryParams,
          $treeDepth,
          $currentDepth + 1,
        );
      }

      $element['is_expanded'] = !empty($element['below']);
      $element['is_collapsed'] = empty($element['below']) && !empty($data['children']);

      $items[$data['id']] = $element;
    }

    return $items;
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheTags(): array {
    $cacheTags = parent::getCacheTags();
    return Cache::mergeTags($cacheTags, [
      'config:' . ContentSettingsForm::CONFIGURATION_NAME,
    ]);
  }

  /**
   * Provides a URL instance for category link.
   *
   * @return \Drupal\Core\Url
   *   The URL instance.
   */
  protected function getLink(): Url {
    return empty($this->configuration['target_page']) ?
      Url::fromRoute('<current>') :
      Url::fromUserInput($this->configuration['target_page']);
  }

}
