<?php

namespace Drupal\commercetools_content\Service;

use Drupal\commercetools\CacheableCommercetoolsResponse;
use Drupal\commercetools\CommercetoolsApiServiceInterface;
use Drupal\commercetools\CommercetoolsProducts;
use Drupal\commercetools\CommercetoolsService;
use Drupal\commercetools\Routing\UiModulesRouteProviderBase;
use Drupal\commercetools_content\Form\CatalogFacetsForm;
use Drupal\commercetools_content\Routing\RouteProvider;
use Drupal\Core\Form\FormBuilderInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Pager\PagerManagerInterface;
use Drupal\Core\Security\TrustedCallbackInterface;
use Drupal\Core\Url;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Provides components for rendering commercetools content.
 */
class CommercetoolsContentComponents implements TrustedCallbackInterface {

  const URL_PARAMS_LIST = [
    'text',
    'filters',
    'queryFilters',
    'facetFilters',
    'sorts',
    'search',
    'facets',
    'category',
  ];

  /**
   * A storage of product list block configurations by the product list index.
   *
   * Used to get the configuration in the filters block and other blocks that
   * should know the current product list filters.
   *
   * @var array
   */
  protected $configurationPerIndex = [];

  /**
   * A cache of the product list results to reuse in different components.
   *
   * @var array
   */
  protected $productListResultsCache = [];

  /**
   * Constructs a CommercetoolsContentComponents object.
   *
   * @param \Drupal\commercetools\CommercetoolsProducts $ctProducts
   *   The Commercetools products service.
   * @param \Drupal\commercetools\CommercetoolsService $ct
   *   The Commercetools service.
   * @param \Drupal\Core\Pager\PagerManagerInterface $pagerManager
   *   The pager manager.
   * @param \Drupal\Core\Form\FormBuilderInterface $formBuilder
   *   The form builder service.
   * @param \Drupal\commercetools\CommercetoolsApiServiceInterface $ctApi
   *   The Commercetools API service.
   * @param \Symfony\Component\HttpFoundation\RequestStack $requestStack
   *   The request stack service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   */
  public function __construct(
    protected CommercetoolsProducts $ctProducts,
    protected CommercetoolsService $ct,
    protected PagerManagerInterface $pagerManager,
    protected FormBuilderInterface $formBuilder,
    protected CommercetoolsApiServiceInterface $ctApi,
    protected RequestStack $requestStack,
    protected MessengerInterface $messenger,
  ) {
  }

  /**
   * {@inheritdoc}
   */
  public static function trustedCallbacks() {
    return ['buildFacets'];
  }

  /**
   * Main method to build the product list.
   */
  public function getProductListComponent(ProductListConfigurationDto $configuration): array {
    // Set the configuration to be able to read them in the blocks.
    $this->configurationPerIndex[$configuration->productListIndex] = $configuration;

    $output = [];
    $products = $this->getProductsForProductList($configuration);
    if (empty($products->getData())) {
      return $output;
    }
    $totalAvailableProducts = $products->getData()['total'] ?? 0;
    $effectiveLimit = isset($configuration->limit) && $configuration->limit > 0
      ? min($configuration->limit, $totalAvailableProducts)
      : $totalAvailableProducts;
    $currentPage = $this->pagerManager->findPage($configuration->productListIndex);
    $facets = (array) $this->getRequestQueryParams($configuration->productListIndex)['facets'] ?? [];
    $output = $this->buildRenderArray($products, $facets, $configuration, $currentPage, $effectiveLimit, $configuration->itemsPerPage);
    $cache = $products->getCacheableMetadata();
    $cache->applyTo($output);
    return $output;
  }

  /**
   * Gets the product list block configuration by the product list index.
   */
  private function getProductListConfiguration(int $productListIndex): ?ProductListConfigurationDto {
    $configuration = $this->configurationPerIndex[$productListIndex] ?? NULL;
    if (!$configuration) {
      // @todo Rework to loading all the available blocks on the current page
      // and search for the product list block with the specified component
      // index.
      $configuration = new ProductListConfigurationDto(
        productListIndex: $productListIndex,
      );
    }
    return $configuration;
  }

  /**
   * Add facet placeholder to build form after building all blocks.
   */
  public function getFacetForm($productListIndex, $targetPage = NULL) {
    return [
      '#pre_render' => [[$this, 'buildFacets']],
      '#product_list_index' => $productListIndex,
      '#target_page' => $targetPage,
    ];
  }

  /**
   * Build facets form.
   */
  public function buildFacets(array $build) {
    $productListIndex = $build['#product_list_index'];
    $productListConfiguration = $this->getProductListConfiguration($productListIndex);
    $result = $this->getProductsForProductList($productListConfiguration);
    $productData = $result->getData();
    $facets = (array) $this->getRequestQueryParams($productListIndex)['facets'] ?? [];
    $args = [
      $productData['facets'] ?? [],
      $facets,
      $productListIndex,
      $build['#target_page'] ?? NULL,
    ];
    $build['form'] = $this->formBuilder->getForm(CatalogFacetsForm::class, ...$args);

    return $build;
  }

  /**
   * Extracts valid query parameters from the request.
   *
   * @param int|null $productListIndex
   *   The product list index. Optional.
   *
   * @return array
   *   All available valid query parameters.
   */
  public function getRequestQueryParams(?int $productListIndex = NULL): array {
    $params = [];
    $request = $this->requestStack->getCurrentRequest();
    $queryParams = $request->query->all();

    if (isset($productListIndex)) {
      foreach (self::URL_PARAMS_LIST as $key) {
        $params[$key] = $queryParams[self::addParameterSuffix($key, $productListIndex)] ?? NULL;
      }
    }
    else {
      // All query parameters if no index provided.
      $params = $queryParams;
    }

    return $params;
  }

  /**
   * Build and execute product list query.
   */
  protected function getProductsForProductList(ProductListConfigurationDto $configuration): CacheableCommercetoolsResponse {
    $filtersFromRequest = $this->getRequestQueryParams($configuration->productListIndex);

    // Check the cached requests to not execute the GraphQL request again.
    $cacheKey = 'commercetools_content:product_list:' . md5(serialize($configuration) . serialize($filtersFromRequest));
    if ($cachedResult = $this->productListResultsCache[$cacheKey] ?? NULL) {
      return $cachedResult;
    }

    $sortValue = ($configuration->sortBy && $configuration->sortOrder)
      ? trim($configuration->sortBy . ' ' . $configuration->sortOrder)
      : NULL;
    $baseQueryArguments = [
      'text' => $filtersFromRequest['search'],
      'facets' => $this->ct->getEnabledFacets(),
      'facetFilters' => $facetFilter ?? [],
      'filters' => $facetFilter ?? [],
      'sorts' => $sortValue,
    ];

    if (isset($filtersFromRequest['facets'])) {
      foreach ($filtersFromRequest['facets'] as $path => $activeFacet) {
        $filterValue = $this->ct->buildFilter($path, array_values($activeFacet));
        $baseQueryArguments['filters'][] = $filterValue;
        $baseQueryArguments['facetFilters'][] = $filterValue;
      }
    }

    // Filter by categories.
    $categoriesList = [];
    // Explicitly set category in component setting has priority.
    if (!empty($configuration->categories)) {
      $categoriesList = array_map('strval', array_unique(array_values($configuration->categories)));
    }
    // If not explicitly set, check dynamic query parameter.
    elseif (!empty($filtersFromRequest['category'])) {
      $categoriesList[] = $filtersFromRequest['category'];
    }
    if ($categoriesList) {
      $baseQueryArguments['filters'][] = $this->ct->buildFilter(
        'categories.id',
        $categoriesList,
        CommercetoolsService::FILTER_TYPE_TREE,
      );
    }

    if (!empty($configuration->skus)) {
      $baseQueryArguments['filters'][] = $this->ct->buildFilter('variants.sku', array_values($configuration->skus));
    }

    if (!empty($configuration->customFilters)) {
      foreach ($configuration->customFilters as $customFilter) {
        $baseQueryArguments['filters'][] = $customFilter;
      }
    }
    $currentPage = $this->pagerManager->findPage($configuration->productListIndex);
    $offset = $currentPage * $configuration->itemsPerPage;

    $queryArguments = array_merge($baseQueryArguments, [
      'limit' => $configuration->itemsPerPage,
      'offset' => $offset,
    ]);

    $result = $this->ctProducts->getProducts($queryArguments);
    $this->productListResultsCache[$cacheKey] = $result;
    return $result;
  }

  /**
   * Builds the render array for products.
   */
  protected function buildRenderArray(
    $productListCacheable,
    array $facets,
    ProductListConfigurationDto $configuration,
    int $currentPage,
    int $effectiveLimit,
    int $itemsPerPage,
  ): array {
    $productData = $productListCacheable->getData();
    $unavailableDataText = $configuration->unavailableDataText;
    $columnClass = $configuration->showFilters ? 'col-lg-4' : 'col-lg-3';
    $totalDisplayableProducts = $effectiveLimit;
    if ($totalDisplayableProducts > 0) {
      $this->pagerManager->createPager($totalDisplayableProducts, $itemsPerPage, $configuration->productListIndex);
    }
    $totalPages = $totalDisplayableProducts > 0
      ? ceil($totalDisplayableProducts / $itemsPerPage)
      : 0;
    $noProductsFound = $totalDisplayableProducts === 0 || ($currentPage >= $totalPages);
    $remainingItems = $effectiveLimit - ($currentPage * $itemsPerPage);
    $maxItemsToDisplay = max(0, min($remainingItems, $itemsPerPage));
    $results = array_slice($productData['results'] ?? [], 0, $maxItemsToDisplay);

    $items = [];
    foreach ($results as $product) {
      if (!$product['slug']) {
        $this->messenger->addError('The product slug value is missing for a product.');
        $url = Url::fromRoute(RouteProvider::ROUTE_PREFIX . UiModulesRouteProviderBase::PAGE_PRODUCT_ROUTE, ['slug' => CommercetoolsProducts::MISSING_SLUG_VALUE]);
      }
      else {
        $url = Url::fromRoute(RouteProvider::ROUTE_PREFIX . UiModulesRouteProviderBase::PAGE_PRODUCT_ROUTE, ['slug' => $product['slug']]);
      }
      $items[] = [
        '#theme' => 'commercetools_product_card',
        '#unavailable_data_text' => $unavailableDataText,
        '#product' => $product,
        '#url' => $url,
        '#attributes' => ['class' => ['card'], 'data-product-list-index' => $configuration->productListIndex],
      ];
    }
    // If no products found and we are not on the first page, throw a 404.
    // @todo Throw only when rendering for the full page, not for the block.
    if (empty($items) && $currentPage > 0) {
      throw new NotFoundHttpException('No products found.');
    }
    $args = [$productData['facets'] ?? [], $facets, $configuration->productListIndex, Url::fromRoute('<current>')];
    return [
      'catalog' => [
        '#theme' => 'commercetools_product_cards',
        '#filters' => $configuration->showFilters ? $this->formBuilder->getForm(CatalogFacetsForm::class, ...$args) : [],
        '#no_products_found' => $noProductsFound,
        '#list' => [
          '#theme' => 'commercetools_product_cards_items',
          '#items' => $items,
          '#item_attributes' => ['class' => ['col-12', 'col-md-6', 'my-2', $columnClass]],
        ],
      ],
      'pager' => !$noProductsFound && $totalDisplayableProducts > 0
        ? [
          '#type' => 'pager',
          '#element' => $configuration->productListIndex,
        ]
        : [],
    ];
  }

  /**
   * Prepares the GET parameter name taking the component index into account.
   *
   * @param string $paramName
   *   The name of parameter.
   * @param int|null $productListIndex
   *   The component index number. Optional.
   *
   * @return string
   *   Component suffix.
   */
  public static function addParameterSuffix(string $paramName, ?int $productListIndex = NULL): string {
    return $paramName . (empty($productListIndex) ? '' : "_{$productListIndex}");
  }

}

/**
 * Defines the configuration for displaying a product list.
 */
class ProductListConfigurationDto {

  public function __construct(
    public ?int $itemsPerPage = NULL,
    public ?string $unavailableDataText = NULL,
    public int $productListIndex = 0,
    public ?bool $showFilters = NULL,
    public ?array $filters = NULL,
    public ?array $facets = NULL,
    public ?string $search = NULL,
    public ?int $limit = NULL,
    public int $offset = 0,
    public ?string $sortBy = NULL,
    public ?string $sortOrder = NULL,
    public ?array $categories = NULL,
    public ?array $skus = NULL,
    public ?array $customFilters = NULL,
  ) {
  }

}
