class CartForm extends HTMLElement {
  constructor() {
    super();
    this.actionHandler = this.actionHandler.bind(this);
    this.maxItems = window.drupalSettings.commercetoolsDecoupled.maxItems;
  }

  connectedCallback() {
    this.renderForm();
  }

  async actionHandler(action, data) {
    window.clearTimeout(this.delay);
    this.delay = window.setTimeout(async () => {
      this.isLoading = true;
      this.renderForm();

      const actions = [];
      switch (action) {
        case 'plus-quantity':
        case 'minus-quantity':
          actions.push({
            changeLineItemQuantity: {
              lineItemId: data.lineItem.id,
              quantity: data.quantity,
            },
          });
          break;

        default:
          actions.push({ removeLineItem: { lineItemId: data.lineItem.id } });
          break;
      }

      this.cart = await window.commercetools.updateCart({
        id: '[current_cart:id]',
        version: '[current_cart:version]',
        actions,
      });
      document.dispatchEvent(
        new CustomEvent('commercetoolsDecoupledCartUpdated', {
          detail: {
            totalItems: this.cart.lineItems.length,
          },
        }),
      );
      this.isLoading = false;
      this.renderForm();
    }, 1000);
  }

  generateSummaryContent() {
    const checkoutPath =
      window.drupalSettings.commercetoolsDecoupled?.checkoutPath;
    if (!this.isSummaryCart) {
      return `
        <div class="col-lg-3 bg-body-tertiary">
          <div class="p-5">
            <h3 class="fw-bold mb-5">${Drupal.t('Summary')}</h3>
            <hr class="my-4">
            <div class="d-flex justify-content-between mb-5">
              <p class='h5'>${Drupal.t('Total price')}</p>
              <p class='h5'>${this.isLoading ? CartForm.placeholder(this.isLoading) : this.cart?.totalPrice?.localizedPrice || ''}</p>
            </div>
            ${
              this.isLoading
                ? ''
                : `
              <a href="${!checkoutPath ? '#' : Drupal.url(checkoutPath.substring(1))}" class="btn btn-primary ${!checkoutPath ? 'disabled' : ''}">${Drupal.t('Order')}</a>
            `
            }
          </div>
        </div>`;
    }
    if (this.cart?.lineItems?.length > 0) {
      return `
        <div class="col-lg-12">
          <div class="mt-3">
            <strong>${Drupal.t('Total price')}:</strong>
            ${this.isLoading ? CartForm.placeholder(this.isLoading) : this.cart.totalPrice.localizedPrice}
          </div>
        </div>`;
    }
    return '';
  }

  renderForm() {
    const summaryContent = this.generateSummaryContent();
    this.innerHTML = `
      <form class="commercetools-content-cart-form">
        <section class="pb-4">
          <div class="overflow-hidden">
            <section class="w-100">
              <div class="row">
                <div class="col-12">
                  <div class="card card-registration card-registration-2 overflow-hidden border-0">
                    <div class="card-body p-0">
                      <div class="row g-0">
                        <div class="${this.isSummaryCart ? 'col-lg-12' : 'col-lg-9'}">
                          <div class="${this.isSummaryCart ? 'pt-2' : 'pt-5 pe-5'}">
                            ${
                              !this.isSummaryCart
                                ? `<div class="d-flex justify-content-between align-items-center mb-5">
                                      <h3 class="fw-bold mb-0">${Drupal.t('Cart items')}</h3>
                                      <h4 class="mb-0 text-muted fs-6">${this.isLoading ? CartForm.placeholder(this.isLoading) : Drupal.formatPlural(this.cart?.totalLineItemQuantity || 0, '1 item', '@count items')}</h4>
                                  </div>`
                                : ''
                            }
                            ${
                              !this.isLoading &&
                              (!this.cart?.lineItems ||
                                this.cart?.lineItems?.length === 0)
                                ? `<div class="cart-empty-text">${Drupal.t('You have no items in your shopping cart.')}</div>`
                                : `<div class="cart-line-items"></div>`
                            }
                          </div>
                        </div>
                        ${summaryContent}
                      </div>
                    </div>
                  </div>
                </div>
              </div>
            </section>
          </div>
        </section>
      </form>
    `;
    if (!!this.cart?.lineItems?.length || this.isLoading) {
      const LineItems = document.createElement('ct-line-items');
      LineItems.lineStyle = 'form';
      LineItems.actionHandler = this.actionHandler;
      LineItems.isLoading = this.isLoading;
      const visibleItems = this.isSummaryCart
        ? this.cart.lineItems.slice(0, this.maxItems)
        : this.cart.lineItems;
      LineItems.lineItems = visibleItems;
      this.querySelector('.cart-line-items').append(LineItems);
    }
  }

  static placeholder() {
    return `<span class="placeholderify"><span style="color: initial;">${Drupal.t('Loading...')}</span></span>`;
  }
}

customElements.define('ct-cart-form', CartForm);
