class CheckoutForm extends HTMLElement {
  async connectedCallback() {
    this.cart = {};
    this.isLoading = true;
    this.renderForm();

    this.cart = await window.commercetools.getCart();
    this.isLoading = false;
    this.renderForm();
  }

  renderForm() {
    const fields = CheckoutForm.addressFields();

    this.innerHTML = `
      <form class="commercetools-content-order-submission ${this.isLoading ? 'placeholderify' : ''}" data-drupal-selector="commercetools-content-order-submission" action="/checkout" method="post" id="commercetools-content-order-submission" accept-charset="UTF-8">
        <fieldset class="shadow p-3 bg-primary bg-opacity-10 form-item form-wrapper shipping-address">
          <div class="fieldset-wrapper">
            <h2 class="pb-3 mb-4 border-bottom">${Drupal.t('Shipping address')}</h2>
             <div class="shipping-address-fields"></div>
          </div>
        </fieldset>

        <div class="mt-4 mb-4 shadow p-3 bg-primary bg-opacity-10">
          <div class="form-item form-type-checkbox form-item-sameAddress">
            <input type="checkbox" id="edit-sameAddress" name="sameAddress" value="1" ${this.cart?.billingAddress ? '' : `checked="checked"`} class="form-checkbox form-check-input">
            <label for="edit-sameAddress">${Drupal.t('Billing address same as shipping address')}</label>
          </div>
        </div>

        <fieldset class="shadow p-3 bg-primary bg-opacity-10 form-item form-wrapper billing-address">
          <div class="fieldset-wrapper">
            <h2 class="pb-3 mb-4 border-bottom">${Drupal.t('Billing address')}</h2>
            <div class="billing-address-fields"></div>
          </div>
        </fieldset>

        <div data-drupal-selector="edit-actions" class="form-actions js-form-wrapper form-wrapper" id="edit-actions">
            <input type="submit" id="edit-submit" name="op" value="${Drupal.t('Save &amp; Continue')}" class="button form-submit btn btn-primary">
        </div>
      </form>
    `;

    ['shipping', 'billing'].forEach((addressType) => {
      const fieldsWr = this.querySelector(`.${addressType}-address-fields`);
      fields.forEach((fieldProperties) => {
        fieldProperties.defaultValue =
          this.cart?.[`${addressType}Address`]?.[fieldProperties.name] ||
          fieldProperties.defaultValue ||
          '';

        const fieldHTML = document.createElement(
          `ct-input-${fieldProperties.type}`,
        );
        fieldHTML.properties = {
          ...fieldProperties,
          name: `${addressType}Address.${fieldProperties.name}`,
        };
        fieldsWr.append(fieldHTML);
      });
    });

    // Display/hide billing address section according to the "Same address" checkbox.
    const sameCheckbox = this.querySelector('input[name="sameAddress"]');
    const checkboxAction = (checked) => {
      const billingAddress = this.querySelector('.billing-address');
      if (checked === true) {
        billingAddress.classList.add('d-none');
        billingAddress
          .querySelectorAll('.required')
          .forEach((el) => el.removeAttribute('required'));
      } else {
        billingAddress.classList.remove('d-none');
        billingAddress
          .querySelectorAll('.required')
          .forEach((el) => el.setAttribute('required', 'required'));
      }
    };
    checkboxAction(sameCheckbox?.checked || false);
    this.querySelector('[name="sameAddress"]').addEventListener('change', (e) =>
      checkboxAction(e.target.checked),
    );

    this.querySelector('form').addEventListener(
      'submit',
      CheckoutForm.submitFormEvent,
    );
  }

  static async submitFormEvent(e) {
    e.preventDefault();

    e.target.classList.add('placeholderify');

    // FormData doesn’t include disabled fields.
    e.target
      .querySelectorAll('input[disabled="disabled"]')
      .forEach((el) => el.removeAttribute('disabled'));

    // Get form values.
    const formData = new FormData(e.target);
    const values = {};
    formData.forEach((v, k) => {
      if (k.includes('.')) {
        const [prefix, fieldName] = k.split('.');
        if (typeof values[prefix] === 'undefined') {
          values[prefix] = {};
        }
        values[prefix][fieldName] = v;
      } else {
        values[k] = v;
      }
    });

    // Add shipping and billing addresses necessary for creating order.
    const actions = [
      {
        setShippingAddress: { address: values.shippingAddress },
      },
      {
        setBillingAddress: {
          address: values.sameAddress
            ? values.shippingAddress
            : values.billingAddress,
        },
      },
    ];

    const cart = await window.commercetools.updateCart({
      id: '[current_cart:id]',
      version: '[current_cart:version]',
      actions,
    });

    // Create order by using cart.
    const order = await window.commercetools.createOrderFromCart({
      id: cart.id || '[current_cart:id]',
      version: cart.version || '[current_cart:version]',
    });

    // Remove cart cookie and redirect to the order page.
    document.cookie = `${drupalSettings.commercetoolsDecoupled.sessionCookieName}=; Max-Age=-1;`;
    const orderPath = `${drupalSettings.commercetoolsDecoupled.orderPathPrefix}/${order.id}`;
    window.location.href = orderPath;
  }

  static addressFields() {
    return [
      {
        label: Drupal.t('First name'),
        type: 'textfield',
        name: 'firstName',
        required: true,
        disabled: false,
      },
      {
        label: Drupal.t('Last name'),
        type: 'textfield',
        name: 'lastName',
        required: true,
        disabled: false,
      },
      {
        label: Drupal.t('Email'),
        type: 'email',
        name: 'email',
        required: true,
      },
      {
        label: Drupal.t('Country'),
        type: 'textfield',
        name: 'country',
        required: true,
        disabled: true,
        defaultValue: window.drupalSettings.commercetoolsDecoupled.priceCountry,
      },
      {
        label: Drupal.t('Street Address'),
        type: 'textfield',
        name: 'streetName',
        required: true,
        disabled: false,
      },
      {
        label: Drupal.t('Street number'),
        type: 'textfield',
        name: 'streetNumber',
        required: false,
        disabled: false,
      },
      {
        label: Drupal.t('State'),
        type: 'textfield',
        name: 'state',
        required: true,
        disabled: false,
      },
      {
        label: Drupal.t('City'),
        type: 'textfield',
        name: 'city',
        required: true,
        disabled: false,
      },
      {
        label: Drupal.t('ZIP code'),
        type: 'textfield',
        name: 'postalCode',
        required: true,
        disabled: false,
      },
      {
        label: Drupal.t('Phone number'),
        type: 'textfield',
        name: 'phone',
        required: true,
        disabled: false,
      },
    ];
  }
}

customElements.define('ct-checkout-form', CheckoutForm);
