class ProductCatalog extends HTMLElement {
  constructor() {
    super();
    this.currentPage = 0;
    this.activeFilters = {};
    this.initialLoad = true;
    this.config = this.loadConfig();

    this.handlePageChange = this.handlePageChange.bind(this);
    this.handleFilterChange = this.handleFilterChange.bind(this);
  }

  loadConfig() {
    const defaultSettings = window.drupalSettings.commercetoolsDecoupled || {};
    return {
      itemsPerPage: this.getAttribute('data-items-per-page')
        ? Number(this.getAttribute('data-items-per-page'))
        : defaultSettings.itemsPerPage,
      categories: this.getAttribute('data-categories')
        ? JSON.parse(this.getAttribute('data-categories'))
        : null,
      skus: this.getAttribute('data-skus')
        ? JSON.parse(this.getAttribute('data-skus'))
        : null,
      customFilters: this.getAttribute('data-custom-filters')
        ? JSON.parse(this.getAttribute('data-custom-filters'))
        : null,
      totalLimit: Number(this.getAttribute('data-total-limit')),
      sortBy: this.getAttribute('data-sort-by'),
      sortOrder: this.getAttribute('data-sort-order'),
      productListIndex: this.getAttribute('data-product-list-index')
        ? Number(this.getAttribute('data-product-list-index'))
        : 0,
      showFilters: this.hasAttribute('data-show-filters')
        ? this.getAttribute('data-show-filters') === 'true'
        : true,
    };
  }

  async connectedCallback() {
    if (!this.config.itemsPerPage || this.config.itemsPerPage <= 0) {
      new Drupal.Message().add(
        Drupal.t('Invalid configuration for items per page.'),
        { type: 'error' },
      );
      return;
    }
    // Placeholder for the products data.
    this.productsData = {
      total: -1,
      products: Array(this.config.itemsPerPage).fill({}),
      facets: [],
    };

    // Loading the real products data.
    const url = new URL(window.location);
    this.currentPage = Number(url.searchParams.get('page')) || 0;
    this.setActiveFilters(url.searchParams);
    await this.updateProductData();
  }

  renderComponent() {
    this.innerHTML = /* html */ `
      <div class="container">
        <div class="row">
          <div class="col-md-${!this.config.showFilters ? '12' : '9'} col-12 ps-0">
            <span class="ct-product-list"></span>
          </div>
          ${
            this.config.showFilters
              ? `
            <div class="col-md-3 col-12 ps-0">
              <span class="ct-facets-form"></span>
            </div>
          `
              : ''
          }
        </div>
        <span class="ct-pager"></span>
      </div>
    `;

    const productList = document.createElement('ct-product-list');
    productList.products = this.productsData.products;
    productList.showFilters = this.config.showFilters;
    productList.isLoading = this.isLoading;
    if (this.config.totalLimit <= 0) {
      this.config.totalLimit = this.productsData.total;
    }
    this.querySelector('.ct-product-list').appendChild(productList);

    if (this.config.showFilters && this.productsData.facets) {
      const facetsForm = document.createElement('ct-facets-form');
      facetsForm.facets = this.productsData.facets;
      facetsForm.activeFilters = this.activeFilters;
      facetsForm.facetChangeHandle = this.handleFilterChange;
      facetsForm.isLoading = this.initialLoad;
      this.querySelector('.ct-facets-form').appendChild(facetsForm);
    }
    if (this.productsData.products.length > 0) {
      const pager = document.createElement('ct-pager');
      const totalPages = Math.ceil(
        Math.min(this.config.totalLimit, this.productsData.total) /
          this.config.itemsPerPage,
      );
      if (totalPages > 1) {
        pager.setPager(this.currentPage, totalPages);
        pager.addEventListener('page-change', this.handlePageChange.bind(this));
        this.querySelector('.ct-pager').appendChild(pager);
      }
    } else if (this.currentPage > 0) {
      // If we're not on the first page - showing that the page does not exist.
      this.innerHTML = '<ct-page-not-found>';
    } else {
      this.innerHTML = '<ct-no-products-found>';
    }
    this.initialLoad = false;
    this.isLoading = false;
  }

  async handlePageChange(event) {
    this.currentPage = event.detail.page;
    const urlParams = new URLSearchParams(window.location.search);
    urlParams.set('page', this.currentPage);
    try {
      await this.updateProductData(urlParams);
    } catch (error) {
      new Drupal.Message().add(
        Drupal.t('An error occurred while changing the page.'),
        {
          type: 'error',
        },
      );
      throw error;
    }
  }

  async handleFilterChange(event) {
    this.currentPage = 0;
    const formData = new FormData(event.currentTarget);
    const params = new URLSearchParams(formData);
    this.setActiveFilters(params);
    await this.updateProductData(params);
  }

  setActiveFilters(params) {
    this.activeFilters = {};
    params.forEach((facetValue, key) => {
      const match = key.match(/^facets\[(.+?)\]\[(.+?)\]$/);
      if (match) {
        const [, path, value] = match;
        this.activeFilters[path] = this.activeFilters[path]
          ? this.activeFilters[path]
          : [];
        this.activeFilters[path].push(value);
      }
    });
  }

  async updateProductData(newSearchParams = null) {
    this.isLoading = true;
    this.renderComponent();

    let limitForThisCall;
    if (this.config.totalLimit > 0) {
      const totalItemsToShow = this.config.totalLimit;
      const maxPageNumber = Math.ceil(
        totalItemsToShow / this.config.itemsPerPage,
      );
      if (this.currentPage >= maxPageNumber) {
        this.productsData.products = [];
        this.isLoading = false;
        this.renderComponent();
        return;
      }
      const remainingItems =
        totalItemsToShow - this.currentPage * this.config.itemsPerPage;
      limitForThisCall = Math.min(this.config.itemsPerPage, remainingItems);
    } else {
      limitForThisCall = this.config.itemsPerPage;
    }
    const activeFilters = this.prepareFilters();
    const productQueryParams = {
      offset: this.currentPage * this.config.itemsPerPage,
      limit: limitForThisCall,
    };
    if (this.config.sortBy || this.config.sortOrder) {
      productQueryParams.sorts = `${this.config.sortBy} ${this.config.sortOrder}`;
    }
    try {
      const response = await window.commercetools.getProducts(
        'productProjectionSearch',
        productQueryParams,
        false,
        activeFilters,
      );
      if (this.config.totalLimit > 0) {
        response.total = Math.min(response.total, this.config.totalLimit);
      }
      this.productsData = response;
    } finally {
      this.isLoading = false;
      this.renderComponent();
      if (newSearchParams) {
        const url = `${drupalSettings.path.currentPath}?${newSearchParams.toString()}`;
        window.history.pushState({ path: url }, '', url);
      }
    }
  }

  prepareFilters() {
    const activeFilters = [];
    // Handle category filters
    if (
      this.config.categories &&
      Object.keys(this.config.categories).length > 0
    ) {
      const categoryValues = [
        ...new Set(
          Object.values(this.config.categories).map((categoryId) =>
            String(categoryId),
          ),
        ),
      ];
      activeFilters.push(
        window.commercetools.buildFilter('categories.id', categoryValues),
      );
    }
    if (this.config.skus && this.config.skus.length > 0) {
      const skuValues = this.config.skus.map((sku) => String(sku));
      activeFilters.push(
        window.commercetools.buildFilter('variants.sku', skuValues),
      );
    }

    if (this.config.customFilters) {
      const conditions = Array.isArray(this.config.customFilters)
        ? this.config.customFilters
        : [this.config.customFilters];
      conditions.forEach((condition) => {
        if (typeof condition === 'string') {
          condition = JSON.parse(condition);
        }
        activeFilters.push(condition);
      });
    }

    // Handle other active filters
    Object.keys(this.activeFilters).forEach((filterName) => {
      const filterValues = this.activeFilters[filterName];
      activeFilters.push(
        window.commercetools.buildFilter(filterName, filterValues),
      );
    });
    return activeFilters;
  }
}

customElements.define('ct-product-catalog', ProductCatalog);
