<?php

namespace Drupal\commercetools;

use Commercetools\Exception\InvalidArgumentException;
use Drupal\Core\Entity\EntityMalformedException;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Site\Settings;
use Drupal\user\UserInterface;
use GraphQL\Actions\Mutation;
use GraphQL\Actions\Query;
use GraphQL\Entities\Node;
use GraphQL\Entities\Variable;

/**
 * The Commercetools customers service.
 */
class CommercetoolsCustomers {

  const BASE_ENTITY_TYPE_ID = 'user';

  const USER_CUSTOMER_ID_FIELD = 'commercetools_customer_id';

  /**
   * CommercetoolsContentCustomers constructor.
   *
   * @param \Drupal\Core\Session\AccountInterface $user
   *   The current user.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\commercetools\CommercetoolsApiServiceInterface $ctApi
   *   The Commercetools API service.
   * @param \Drupal\Core\Site\Settings $settings
   *   The settings instance.
   */
  public function __construct(
    protected readonly AccountInterface $user,
    protected readonly EntityTypeManagerInterface $entityTypeManager,
    protected readonly CommercetoolsApiServiceInterface $ctApi,
    protected readonly Settings $settings,
  ) {
  }

  /**
   * Provides a customer ID field storage definitions.
   *
   * @return \Drupal\Core\Field\FieldStorageDefinitionInterface
   *   The customer ID field storage definitions.
   */
  public static function customerIdFieldDefinitions(): FieldStorageDefinitionInterface {
    return BaseFieldDefinition::create('string')
      ->setLabel('commercetools customer ID')
      ->setDescription('External commercetools customer ID for internal use.')
      ->setRevisionable(FALSE)
      ->setSetting('max_length', 86);
  }

  /**
   * Loads a Commercetools customer by ID.
   *
   * @param string $customerId
   *   The Commercetools customer ID.
   *
   * @return array
   *   An existing customer data from API.
   *
   * @throws \Throwable
   */
  public function load(string $customerId): array {
    $arguments = [
      'id' => new Variable('id', 'String!'),
    ];
    $query = new Query('customer', $arguments);
    $this->addFields($query);
    $variables = [
      'id' => $customerId,
    ];
    $response = $this->ctApi->executeGraphQlOperation($query, $variables);
    return (array) $response->getData()['customer'];
  }

  /**
   * Creates a new Commercetools customer.
   *
   * @param array $draft
   *   The array of values which will be used for creating customer.
   *
   * @return array
   *   An added customer data from API.
   *
   * @throws \Commercetools\Exception\InvalidArgumentException
   * @throws \Throwable
   */
  public function create(array $draft = []): array {
    // Force password to be non-mandatory.
    $draft['authenticationMode'] = 'ExternalAuth';

    // Set current user's email if not provided.
    if (empty($draft['email'])) {
      $email = $this->getUserEmail();
      if ($email) {
        $draft['email'] = $email;
      }
      else {
        throw new InvalidArgumentException('An email is required to create a customer');
      }
    }

    $arguments = [
      'draft' => new Variable('draft', 'CustomerSignUpDraft!'),
    ];
    $mutation = new Mutation('customerSignUp', $arguments);
    $customer = $mutation->customer([]);
    $this->addFields($customer);
    $variables = [
      'draft' => $draft,
    ];
    $response = $this->ctApi->executeGraphQlOperation($mutation, $variables);
    return (array) $response->getData()['customerSignUp'];
  }

  /**
   * Deletes a Commercetools customer.
   *
   * @param string $customerId
   *   The Commercetools customer ID.
   * @param int $version
   *   The last seen version of the Commercetools customer.
   *
   * @return array
   *   A customer data being deleted.
   *
   * @throws \Throwable
   */
  public function delete(string $customerId, int $version): array {
    $arguments = [
      'id' => new Variable('id', 'String'),
      'version' => new Variable('version', 'Long!'),
    ];
    $mutation = new Mutation('deleteCustomer', $arguments);
    $this->addFields($mutation);
    $variables = [
      'id' => $customerId,
      'version' => $version,
    ];
    $response = $this->ctApi->executeGraphQlOperation($mutation, $variables);
    return (array) $response->getData()['deleteCustomer'];
  }

  /**
   * Retrieves a Commercetools customer for the given or current user.
   *
   * Conditionally store Customer ID into user on update.
   *
   * @param \Drupal\user\UserInterface|null $user
   *   The user to retrieve a customer by.
   *   Fallbacks to current user if not provided.
   * @param bool $seek
   *   Indicates whether to seek for existing customer by email.
   *   Updates the user entity with customer ID if found.
   * @param bool $create
   *   Indicates whether to create a new customer.
   *   Updates the user entity with created customer ID.
   * @param bool $store
   *   Indicates whether to store Customer ID into user.
   *
   * @return array|null
   *   A customer data retrieved.
   *
   * @throws \Throwable
   */
  public function getCustomerByUser(?UserInterface $user = NULL, bool $seek = TRUE, bool $create = TRUE, bool $store = TRUE): ?array {
    // First, try to load Customer by stored ID.
    $id = $this->getUserCustomerId($user);
    if ($id) {
      $data = $this->load($id);
      if (!empty($data['version'])) {
        return $data;
      }
    }

    $email = $this->getUserEmail($user);

    // Second, conditionally seek for Customer by email.
    if ($email && $seek) {
      $data = $this->query(['email' => $email]);
      $customer = reset($data);
      if (!empty($customer['version'])) {
        if ($store) {
          $this->setUserCustomerId($customer['id'], $user);
        }
        return $customer;
      }
    }

    // Last, conditionally create a Customer.
    if ($email && $create) {
      $data = $this->create(['email' => $email]);
      if (!empty($data['version'])) {
        if ($store) {
          $this->setUserCustomerId($data['id'], $user);
        }
        return $data;
      }
    }

    return NULL;
  }

  /**
   * Removes a Commercetools customer for the given or current user.
   *
   * @param \Drupal\user\UserInterface|null $user
   *   The user to retrieve by.
   *   Fallbacks to current user if not provided.
   *
   * @return array|null
   *   Customer data being deleted.
   *
   * @throws \Throwable
   */
  public function deleteCustomerByUser(?UserInterface $user = NULL): ?array {
    /** @var \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher */
    $id = $this->getUserCustomerId($user);
    $version = NULL;
    if ($id) {
      $customer = $this->load($id);
      $version = $customer['version'] ?? NULL;
    }
    return $id && $version ? $this->delete($id, $version) : NULL;
  }

  /**
   * Sets the user's customer ID.
   *
   * @param string $customerId
   *   The Commercetools customer ID.
   * @param \Drupal\user\UserInterface|null $user
   *   The user entity to update. Fallbacks to current user if not provided.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  public function setUserCustomerId(string $customerId, ?UserInterface $user = NULL): void {
    if (empty($user)) {
      $user = $this->getCurrentUser();
    }
    if (!$user->hasField(self::USER_CUSTOMER_ID_FIELD)) {
      throw new EntityMalformedException(self::USER_CUSTOMER_ID_FIELD . ' is missing on ' . self::BASE_ENTITY_TYPE_ID);
    }
    if ($user->id()) {
      $user->set(self::USER_CUSTOMER_ID_FIELD, $customerId);
      $user->save();
    }
  }

  /**
   * Returns the user's customer ID if set.
   *
   * @param \Drupal\user\UserInterface|null $user
   *   The user to retrieve by. Fallbacks to current user if not provided.
   *
   * @return ?string
   *   A user customer ID if available.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  private function getUserCustomerId(?UserInterface $user = NULL): ?string {
    if (empty($user)) {
      $user = $this->getCurrentUser();
    }
    if (!$user->hasField(self::USER_CUSTOMER_ID_FIELD)) {
      throw new EntityMalformedException(self::USER_CUSTOMER_ID_FIELD . ' is missing on ' . self::BASE_ENTITY_TYPE_ID);
    }

    if ($user->get(self::USER_CUSTOMER_ID_FIELD)->isEmpty()) {
      return NULL;
    }
    else {
      return $user->get(self::USER_CUSTOMER_ID_FIELD)->value;
    }
  }

  /**
   * Returns the current user's customer ID if available.
   *
   * @return ?string
   *   A customer ID if available.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Throwable
   */
  public function getCurrentUserCustomerId(): ?string {
    // First, quickly check if customer ID is set for a user account.
    $customerId = $this->getUserCustomerId();
    // If no, try to determine a customer ID via API call.
    if (!$customerId) {
      if ($customer = $this->getCustomerByUser()) {
        $customerId = $customer['id'];
      }
    }

    return $customerId;
  }

  /**
   * Removes all customer IDs previously assigned to Users.
   *
   * @param array $uids
   *   The User IDs.
   * @param int $batchSize
   *   The number of entities to process at once.
   */
  public function unsetCustomerIdForUsers(array $uids = [], int $batchSize = 0): void {
    $storage = $this->entityTypeManager
      ->getStorage(self::BASE_ENTITY_TYPE_ID);

    if (!$uids) {
      $uids = $storage->getQuery()
        ->accessCheck(FALSE)
        ->exists(self::USER_CUSTOMER_ID_FIELD)
        ->execute();
    }

    // Nothing to clean up.
    if (!$uids) {
      return;
    }

    if (!$batchSize) {
      $batchSize = $this->settings->get('entity_update_batch_size', 50);
    }

    $chunks = array_chunk($uids, $batchSize);
    foreach ($chunks as $ids) {
      $users = $storage->loadMultiple($ids);
      /** @var \Drupal\user\UserInterface $user */
      foreach ($users as $user) {
        $user->set(self::USER_CUSTOMER_ID_FIELD, NULL)
          ->save();
      }
      // Prevent memory leakage for numerous entities.
      $storage->resetCache($ids);
    }
  }

  /**
   * Adds customer fields to request node.
   *
   * @param \GraphQL\Entities\Node $query
   *   The node to which the fields will be added.
   */
  protected function addFields(Node $query): void {
    $query->use('id', 'version', 'email');
  }

  /**
   * Query a Commercetools customers.
   *
   * @param array $where
   *   The where clause array.
   *
   * @return array
   *   A customers dataset queried from API.
   *
   * @throws \Throwable
   */
  private function query(array $where): array {
    $arguments = [
      'limit' => 1,
      'where' => new Variable('where', 'String'),
    ];
    $query = new Query('customers', $arguments);
    $result = $query->results([]);
    $this->addFields($result);
    $variables = [
      'where' => CommercetoolsService::whereToString($where),
    ];
    $response = $this->ctApi->executeGraphQlOperation($query, $variables);
    return (array) $response->getData()['customers']['results'];
  }

  /**
   * Loads current user entity.
   *
   * @return \Drupal\user\UserInterface
   *   The current user entity.
   */
  private function getCurrentUser(): UserInterface {
    /** @var \Drupal\user\UserInterface $user */
    $user = $this->entityTypeManager->getStorage('user')
      ->load($this->user->id());
    return $user;
  }

  /**
   * Returns email of given or currently logged-in user if any.
   *
   * @param \Drupal\user\UserInterface|null $user
   *   The user entity.
   *
   * @return string|null
   *   An email address if any.
   */
  private function getUserEmail(?UserInterface $user = NULL): ?string {
    if ($user) {
      return $user->getEmail();
    }
    elseif ($this->user->isAuthenticated()) {
      return $this->user->getEmail();
    }

    return NULL;
  }

}
