<?php

namespace Drupal\commercetools;

use Drupal\block\Entity\Block;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Entity\EntityDefinitionUpdateManagerInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drush\Drush;

/**
 * Contains implementations of Commercetools historical updates.
 */
class CommercetoolsUpdates {

  /**
   * Installs or re-installs a customer ID field.
   *
   * @return int
   *   An operation result code.
   */
  public static function installCustomerIdField(): int {
    $updateManager = \Drupal::entityDefinitionUpdateManager();

    // Uninstall field if present.
    $existingField = $updateManager->getFieldStorageDefinition(
      CommercetoolsCustomers::USER_CUSTOMER_ID_FIELD,
      CommercetoolsCustomers::BASE_ENTITY_TYPE_ID,
    );
    if ($existingField) {
      $updateManager->uninstallFieldStorageDefinition($existingField);
    }

    $updateManager->installFieldStorageDefinition(
      CommercetoolsCustomers::USER_CUSTOMER_ID_FIELD,
      CommercetoolsCustomers::BASE_ENTITY_TYPE_ID,
      'commercetools',
      CommercetoolsCustomers::customerIdFieldDefinitions(),
    );

    return $existingField ?
      EntityDefinitionUpdateManagerInterface::DEFINITION_UPDATED :
      EntityDefinitionUpdateManagerInterface::DEFINITION_CREATED;
  }

  /**
   * Updates the product attributes configuration key.
   */
  public static function moveProductAttributesConfig(): void {
    $configFactory = \Drupal::configFactory();
    $config = $configFactory->getEditable(CommercetoolsService::CONFIGURATION_NAME);

    $configuration = [];
    if ($oldConfiguration = $config->get('product_filter_attributes')) {
      foreach (array_filter($oldConfiguration) as $attrName => $enable) {
        $configuration[$attrName] = [
          'use_on_page' => TRUE,
          'use_as_facet' => TRUE,
        ];
      }
    }
    $config->clear('product_filter_attributes');
    $config->set('product_attributes', $configuration);
    $config->save();
  }

  /**
   * Sets the initial items_per_page configuration value.
   */
  public static function setItemsPerPageConfiguration(): void {
    $config_factory = \Drupal::configFactory();
    $config = $config_factory->getEditable(CommercetoolsService::CONFIGURATION_NAME);
    $config->set(CommercetoolsService::CONFIG_ITEMS_PER_PAGE, 9);
    $config->save();
  }

  /**
   * Sets the initial items_per_page configuration value.
   */
  public static function setUnavailableDataTextConfiguration(): void {
    $config_factory = \Drupal::configFactory();
    $config = $config_factory->getEditable(CommercetoolsService::CONFIGURATION_NAME);
    $config->set(CommercetoolsService::CONFIG_UNAVAILABLE_DATA_TEXT, 'N/A');
    $config->save();
  }

  /**
   * Sets the initial translation configuration.
   */
  public static function setLanguageFallbacksConfiguration(): void {
    // Languages fallbacks are deprecated in 1.9 in favour of localization.
  }

  /**
   * Turns on the response caching in configuration.
   */
  public static function enableResponseCache(): void {
    self::setConfigKeys([
      CommercetoolsApiServiceInterface::CONFIG_CACHE_RESPONSES_TTL => Cache::PERMANENT,
    ], CommercetoolsApiServiceInterface::CONFIGURATION_NAME);
  }

  /**
   * Saves the new values to the configuration keys.
   *
   * @param array $values
   *   A key-value array of values.
   * @param string $name
   *   The configuration key name. The commercetools settings key by default.
   */
  public static function setConfigKeys(
    array $values,
    string $name = CommercetoolsService::CONFIGURATION_NAME,
  ): void {
    $config_factory = \Drupal::configFactory();
    $config = $config_factory->getEditable($name);
    foreach ($values as $key => $value) {
      $config->set($key, $value);
    }
    $config->save(TRUE);
  }

  /**
   * Unsets specific values in a configuration.
   *
   * @param array $keys
   *   The list of configuration option keys to unset.
   * @param string $name
   *   The configuration key name. The commercetools settings key by default.
   */
  public static function clearConfigKeys(
    array $keys,
    string $name = CommercetoolsService::CONFIGURATION_NAME,
  ): void {
    $config_factory = \Drupal::configFactory();
    $config = $config_factory->getEditable($name);
    foreach ($keys as $key) {
      $config->clear($key);
    }
    $config->save(TRUE);
  }

  /**
   * Moves configuration values from one key to another.
   *
   * @param array $fromTo
   *   An array of the config names and keys, with the source key as the key
   *   and the target key as the value.
   *   Example:
   *   - `["my_old_config:my_old_key" => "my_new_config:my_new_key"]`.
   */
  public static function migrateConfigKeys(
    array $fromTo,
  ): void {
    $configFactory = \Drupal::configFactory();
    $configs = [];
    foreach ($fromTo as $from => $to) {
      [$fromName, $fromKey] = explode(':', $from);
      [$toName, $toKey] = explode(':', $to);
      $configs[$fromName] ??= $configFactory->getEditable($fromName);
      $configs[$toName] ??= $configFactory->getEditable($toName);
      $configs[$toName]->set($toKey, $configs[$fromName]->get($fromKey));
      $configs[$fromName]->clear($fromKey);
    }
    foreach ($configs as $config) {
      $config->save(TRUE);
    }
  }

  /**
   * Migrate the commercetools API configuration to a separate config.
   */
  public static function update10009MigrateConfigToApi() {
    CommercetoolsUpdates::migrateConfigKeys([
      'commercetools.settings:client_id' => 'commercetools.api:client_id',
      'commercetools.settings:client_secret' => 'commercetools.api:client_secret',
      'commercetools.settings:project_key' => 'commercetools.api:project_key',
      'commercetools.settings:scope' => 'commercetools.api:scope',
      'commercetools.settings:api_proxy_path' => 'commercetools.api:api_proxy_path',
      'commercetools.settings:cache_responses_disabled' => 'commercetools.api:cache_responses_disabled',
    ]);

    $configFactory = \Drupal::configFactory();
    $apiUrl = $configFactory->get('commercetools.settings')->get('api_url');
    $url = parse_url($apiUrl);
    $parts = explode('.', $url['host']);
    $hostedRegion = $parts[0] . '.' . $parts[1];
    $apiUrl = $configFactory->getEditable('commercetools.api')
      ->set('hosted_region', $hostedRegion)
      ->save();
    $configFactory->getEditable('commercetools.settings')
      ->clear('auth_url')
      ->clear('api_url')
      ->clear('session_api_url')
      ->save();
    return t('commercetools commercetools API configuration has been migrated');
  }

  /**
   * Displays a message to the user using Drush or Drupal messenger.
   *
   * @param \Drupal\Core\StringTranslation\TranslatableMarkup $message
   *   The message to display.
   */
  public static function displayMessage(TranslatableMarkup $message): void {
    if (class_exists(Drush::class) && Drush::hasContainer()) {
      Drush::logger()->notice($message);
    }
    else {
      \Drupal::messenger()->addStatus($message);
    }
  }

  /**
   * Update block settings field "component_index" to "product_list_index".
   */
  public static function updateBlockSettingsProductListIndex(array $blocks_plugins_ids): void {
    $block_names = \Drupal::entityQuery('block')
      ->condition('plugin', $blocks_plugins_ids)
      ->execute();

    if (!$block_names) {
      return;
    }

    foreach ($block_names as $block_name) {
      $block = Block::load($block_name);
      $block_settings = $block->get('settings');
      if (!is_array($block_settings) || !array_key_exists('component_index', $block_settings)) {
        continue;
      }
      $block_settings['product_list_index'] = $block_settings['component_index'];
      unset($block_settings['component_index']);
      $block->set('settings', $block_settings);
      $block->save();
    }
  }

}
