<?php

namespace Drupal\commercetools\Plugin\Block;

use Drupal\Core\Cache\Cache;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines a base class for listing blocks of commercetools categories.
 */
abstract class CommercetoolsCategoriesListBlockBase extends CommercetoolsCatalogActionBlockBase {

  const MAXIMUM_EXPANDED_TREE = 20;

  /**
   * The Commercetools service.
   *
   * @var \Drupal\commercetools\CommercetoolsService
   */
  protected $ct;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): self {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->ct = $container->get('commercetools');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getBlockConfigKeys(): array {
    $keys = parent::getBlockConfigKeys();
    $keys[] = 'max_level';
    return $keys;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'max_level' => 1,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function blockForm($form, FormStateInterface $form_state): array {
    $form = parent::blockForm($form, $form_state);

    $categories = $this->ct->getProductCategoriesTree();
    $maxDepth = $this->getCategoriesTreeDepth($categories);
    $options = range(0, $maxDepth);
    $options[0] = $this->t('Unlimited');
    $form['max_level'] = [
      '#type' => 'select',
      '#title' => $this->t('Number of levels to display'),
      '#default_value' => $this->configuration['max_level'],
      '#options' => $options,
      '#description' => $this->t('Currently active category always expands its children. Sibling categories of level that exceeds this number will be hidden from display.'),
      '#required' => TRUE,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  abstract public function build(): array;

  /**
   * {@inheritdoc}
   */
  public function getCacheContexts(): array {
    $cacheContexts = parent::getCacheContexts();
    return Cache::mergeContexts($cacheContexts, [
      'url.query_args',
    ]);
  }

  /**
   * Provides a categories tree.
   *
   * @param string|null $activeCategory
   *   Any category ID to build active-trail on. Optional.
   *
   * @return array
   *   A tree of categories.
   */
  protected function getCategoriesTree(?string $activeCategory = NULL): array {
    $categories = $this->ct->getProductCategories();

    // Set active-trail value if available.
    if ($activeCategory) {
      do {
        $categories[$activeCategory]['in_active_trail'] = TRUE;
        $activeCategory = $categories[$activeCategory]['parent'] ?? NULL;
      } while (array_key_exists($activeCategory, $categories));
    }

    return $this->ct->getProductCategoriesTree($categories);
  }

  /**
   * Returns the maximum depth of categories tree.
   *
   * @param array $list
   *   A data structure representing the categories tree.
   *
   * @return int
   *   The maximum array depth.
   */
  protected function getCategoriesTreeDepth(array $list): int {
    $max_indentation = 1;
    $array_str = print_r($list, TRUE);
    $lines = explode("\n", $array_str);
    foreach ($lines as $line) {
      if (!stripos($line, '=> Array')) {
        continue;
      }
      $indentation = (strlen($line) - strlen(ltrim($line))) / 4;
      if ($indentation > $max_indentation) {
        $max_indentation = $indentation;
      }
    }
    return (int) ceil(($max_indentation - 1) / 4);
  }

}
